"use strict";
/**
 * Copyright 2023 actions-toolkit authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Install = void 0;
const crypto_1 = __importDefault(require("crypto"));
const fs_1 = __importDefault(require("fs"));
const os_1 = __importDefault(require("os"));
const path_1 = __importDefault(require("path"));
const core = __importStar(require("@actions/core"));
const httpm = __importStar(require("@actions/http-client"));
const tc = __importStar(require("@actions/tool-cache"));
const cache = __importStar(require("@actions/cache"));
const semver = __importStar(require("semver"));
const util = __importStar(require("util"));
const buildx_1 = require("./buildx");
const context_1 = require("../context");
const exec_1 = require("../exec");
const docker_1 = require("../docker/docker");
const git_1 = require("../git");
const util_1 = require("../util");
class Install {
    _standalone;
    constructor(opts) {
        this._standalone = opts?.standalone;
    }
    /*
     * Download buildx binary from GitHub release
     * @param version semver version or latest
     * @returns path to the buildx binary
     */
    async download(v) {
        const version = await Install.getDownloadVersion(v);
        core.debug(`Install.download version: ${version.version}`);
        const release = await Install.getRelease(version);
        core.debug(`Install.download release tag name: ${release.tag_name}`);
        const vspec = await this.vspec(release.tag_name);
        core.debug(`Install.download vspec: ${vspec}`);
        const c = semver.clean(vspec) || '';
        if (!semver.valid(c)) {
            throw new Error(`Invalid Buildx version "${vspec}".`);
        }
        const installCache = new InstallCache(version.key != 'official' ? `buildx-dl-bin-${version.key}` : 'buildx-dl-bin', vspec);
        const cacheFoundPath = await installCache.find();
        if (cacheFoundPath) {
            core.info(`Buildx binary found in ${cacheFoundPath}`);
            return cacheFoundPath;
        }
        const downloadURL = util.format(version.downloadURL, vspec, this.filename(vspec));
        core.info(`Downloading ${downloadURL}`);
        const htcDownloadPath = await tc.downloadTool(downloadURL);
        core.debug(`Install.download htcDownloadPath: ${htcDownloadPath}`);
        const cacheSavePath = await installCache.save(htcDownloadPath);
        core.info(`Cached to ${cacheSavePath}`);
        return cacheSavePath;
    }
    /*
     * Build buildx binary from source
     * @param gitContext git repo context
     * @returns path to the buildx binary
     */
    async build(gitContext) {
        const vspec = await this.vspec(gitContext);
        core.debug(`Install.build vspec: ${vspec}`);
        const installCache = new InstallCache('buildx-build-bin', vspec);
        const cacheFoundPath = await installCache.find();
        if (cacheFoundPath) {
            core.info(`Buildx binary found in ${cacheFoundPath}`);
            return cacheFoundPath;
        }
        const outputDir = path_1.default.join(context_1.Context.tmpDir(), 'buildx-build-cache');
        const buildCmd = await this.buildCommand(gitContext, outputDir);
        const buildBinPath = await exec_1.Exec.getExecOutput(buildCmd.command, buildCmd.args, {
            ignoreReturnCode: true
        }).then(res => {
            if (res.stderr.length > 0 && res.exitCode != 0) {
                throw new Error(`build failed with: ${res.stderr.match(/(.*)\s*$/)?.[0]?.trim() ?? 'unknown error'}`);
            }
            return `${outputDir}/buildx`;
        });
        const cacheSavePath = await installCache.save(buildBinPath);
        core.info(`Cached to ${cacheSavePath}`);
        return cacheSavePath;
    }
    async installStandalone(binPath, dest) {
        core.info('Standalone mode');
        dest = dest || context_1.Context.tmpDir();
        const binDir = path_1.default.join(dest, 'buildx-bin-standalone');
        if (!fs_1.default.existsSync(binDir)) {
            fs_1.default.mkdirSync(binDir, { recursive: true });
        }
        const binName = os_1.default.platform() == 'win32' ? 'buildx.exe' : 'buildx';
        const buildxPath = path_1.default.join(binDir, binName);
        fs_1.default.copyFileSync(binPath, buildxPath);
        core.info('Fixing perms');
        fs_1.default.chmodSync(buildxPath, '0755');
        core.addPath(binDir);
        core.info('Added Buildx to PATH');
        core.info(`Binary path: ${buildxPath}`);
        return buildxPath;
    }
    async installPlugin(binPath, dest) {
        core.info('Docker plugin mode');
        dest = dest || docker_1.Docker.configDir;
        const pluginsDir = path_1.default.join(dest, 'cli-plugins');
        if (!fs_1.default.existsSync(pluginsDir)) {
            fs_1.default.mkdirSync(pluginsDir, { recursive: true });
        }
        const binName = os_1.default.platform() == 'win32' ? 'docker-buildx.exe' : 'docker-buildx';
        const pluginPath = path_1.default.join(pluginsDir, binName);
        fs_1.default.copyFileSync(binPath, pluginPath);
        core.info('Fixing perms');
        fs_1.default.chmodSync(pluginPath, '0755');
        core.info(`Plugin path: ${pluginPath}`);
        return pluginPath;
    }
    async buildCommand(gitContext, outputDir) {
        const buildxStandaloneFound = await new buildx_1.Buildx({ standalone: true }).isAvailable();
        const buildxPluginFound = await new buildx_1.Buildx({ standalone: false }).isAvailable();
        let buildStandalone = false;
        if ((await this.isStandalone()) && buildxStandaloneFound) {
            core.debug(`Install.buildCommand: Buildx standalone found, build with it`);
            buildStandalone = true;
        }
        else if (!(await this.isStandalone()) && buildxPluginFound) {
            core.debug(`Install.buildCommand: Buildx plugin found, build with it`);
            buildStandalone = false;
        }
        else if (buildxStandaloneFound) {
            core.debug(`Install.buildCommand: Buildx plugin not found, but standalone found so trying to build with it`);
            buildStandalone = true;
        }
        else if (buildxPluginFound) {
            core.debug(`Install.buildCommand: Buildx standalone not found, but plugin found so trying to build with it`);
            buildStandalone = false;
        }
        else {
            throw new Error(`Neither buildx standalone or plugin have been found to build from ref ${gitContext}`);
        }
        const args = ['build', '--target', 'binaries', '--platform', 'local', '--build-arg', 'BUILDKIT_CONTEXT_KEEP_GIT_DIR=1', '--output', `type=local,dest=${outputDir}`];
        if (process.env.GIT_AUTH_TOKEN) {
            args.push('--secret', 'id=GIT_AUTH_TOKEN');
        }
        args.push(gitContext);
        //prettier-ignore
        return await new buildx_1.Buildx({ standalone: buildStandalone }).getCommand(args);
    }
    async isStandalone() {
        const standalone = this._standalone ?? !(await docker_1.Docker.isAvailable());
        core.debug(`Install.isStandalone: ${standalone}`);
        return standalone;
    }
    filename(version) {
        let arch;
        switch (os_1.default.arch()) {
            case 'x64': {
                arch = 'amd64';
                break;
            }
            case 'ppc64': {
                arch = 'ppc64le';
                break;
            }
            case 'arm': {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                const arm_version = process.config.variables.arm_version;
                arch = arm_version ? 'arm-v' + arm_version : 'arm';
                break;
            }
            default: {
                arch = os_1.default.arch();
                break;
            }
        }
        const platform = os_1.default.platform() == 'win32' ? 'windows' : os_1.default.platform();
        const ext = os_1.default.platform() == 'win32' ? '.exe' : '';
        return util.format('buildx-v%s.%s-%s%s', version, platform, arch, ext);
    }
    /*
     * Get version spec (fingerprint) for cache key. If versionOrRef is a valid
     * Git context, then return the SHA of the ref along the repo and owner and
     * create a hash of it. Otherwise, return the versionOrRef (semver) as is
     * without the 'v' prefix.
     */
    async vspec(versionOrRef) {
        if (!util_1.Util.isValidRef(versionOrRef)) {
            const v = versionOrRef.replace(/^v+|v+$/g, '');
            core.info(`Use ${v} version spec cache key for ${versionOrRef}`);
            return v;
        }
        // eslint-disable-next-line prefer-const
        let [baseURL, ref] = versionOrRef.split('#');
        if (ref.length == 0) {
            ref = 'master';
        }
        let sha;
        if (ref.match(/^[0-9a-fA-F]{40}$/)) {
            sha = ref;
        }
        else {
            sha = await git_1.Git.remoteSha(baseURL, ref, process.env.GIT_AUTH_TOKEN);
        }
        const [owner, repo] = baseURL.substring('https://github.com/'.length).split('/');
        const key = `${owner}/${util_1.Util.trimSuffix(repo, '.git')}/${sha}`;
        const hash = crypto_1.default.createHash('sha256').update(key).digest('hex');
        core.info(`Use ${hash} version spec cache key for ${key}`);
        return hash;
    }
    static async getDownloadVersion(v) {
        let [repoKey, version] = v.split(':');
        if (!version) {
            version = repoKey;
            repoKey = 'official';
        }
        switch (repoKey) {
            case 'official': {
                return {
                    key: repoKey,
                    version: version,
                    downloadURL: 'https://github.com/docker/buildx/releases/download/v%s/%s',
                    releasesURL: 'https://raw.githubusercontent.com/docker/actions-toolkit/main/.github/buildx-releases.json'
                };
            }
            case 'lab': {
                return {
                    key: repoKey,
                    version: version,
                    downloadURL: 'https://github.com/docker/buildx-desktop/releases/download/v%s/%s',
                    releasesURL: 'https://raw.githubusercontent.com/docker/actions-toolkit/main/.github/buildx-lab-releases.json'
                };
            }
            default: {
                throw new Error(`Cannot find buildx version for ${v}`);
            }
        }
    }
    static async getRelease(version) {
        const http = new httpm.HttpClient('docker-actions-toolkit');
        const resp = await http.get(version.releasesURL);
        const body = await resp.readBody();
        const statusCode = resp.message.statusCode || 500;
        if (statusCode >= 400) {
            throw new Error(`Failed to get Buildx releases from ${version.releasesURL} with status code ${statusCode}: ${body}`);
        }
        const releases = JSON.parse(body);
        if (!releases[version.version]) {
            throw new Error(`Cannot find Buildx release ${version.version} in ${version.releasesURL}`);
        }
        return releases[version.version];
    }
}
exports.Install = Install;
class InstallCache {
    htcName;
    htcVersion;
    ghaCacheKey;
    cacheDir;
    cacheFile;
    cachePath;
    constructor(htcName, htcVersion) {
        this.htcName = htcName;
        this.htcVersion = htcVersion;
        this.ghaCacheKey = util.format('%s-%s-%s', this.htcName, this.htcVersion, this.platform());
        this.cacheDir = path_1.default.join(buildx_1.Buildx.configDir, '.bin', htcVersion, this.platform());
        this.cacheFile = os_1.default.platform() == 'win32' ? 'docker-buildx.exe' : 'docker-buildx';
        this.cachePath = path_1.default.join(this.cacheDir, this.cacheFile);
        if (!fs_1.default.existsSync(this.cacheDir)) {
            fs_1.default.mkdirSync(this.cacheDir, { recursive: true });
        }
    }
    async save(file) {
        core.debug(`InstallCache.save ${file}`);
        const cachePath = this.copyToCache(file);
        const htcPath = await tc.cacheDir(this.cacheDir, this.htcName, this.htcVersion, this.platform());
        core.debug(`InstallCache.save cached to hosted tool cache ${htcPath}`);
        if (cache.isFeatureAvailable()) {
            core.debug(`InstallCache.save caching ${this.ghaCacheKey} to GitHub Actions cache`);
            await cache.saveCache([this.cacheDir], this.ghaCacheKey);
        }
        return cachePath;
    }
    async find() {
        let htcPath = tc.find(this.htcName, this.htcVersion, this.platform());
        if (htcPath) {
            core.info(`Restored from hosted tool cache ${htcPath}`);
            return this.copyToCache(`${htcPath}/${this.cacheFile}`);
        }
        if (cache.isFeatureAvailable()) {
            core.debug(`GitHub Actions cache feature available`);
            if (await cache.restoreCache([this.cacheDir], this.ghaCacheKey)) {
                core.info(`Restored ${this.ghaCacheKey} from GitHub Actions cache`);
                htcPath = await tc.cacheDir(this.cacheDir, this.htcName, this.htcVersion, this.platform());
                core.info(`Restored to hosted tool cache ${htcPath}`);
                return this.copyToCache(`${htcPath}/${this.cacheFile}`);
            }
        }
        else {
            core.info(`GitHub Actions cache feature not available`);
        }
        return '';
    }
    copyToCache(file) {
        core.debug(`Copying ${file} to ${this.cachePath}`);
        fs_1.default.copyFileSync(file, this.cachePath);
        fs_1.default.chmodSync(this.cachePath, '0755');
        return this.cachePath;
    }
    platform() {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const arm_version = process.config.variables.arm_version;
        return `${os_1.default.platform()}-${os_1.default.arch()}${arm_version ? 'v' + arm_version : ''}`;
    }
}
//# sourceMappingURL=install.js.map